# frozen_string_literal: true

module Heimdall
  # Container for any uniquely identifiable object
  # @abstract
  class UUIDObject
    UUIDSYM = (('0'..'9').to_a + ('a'..'f').to_a).freeze
    class << self
      # Add uuids to subclass
      def inherited(subclass)
        @uuids ||= {}
        @foreign_ids ||= {}
        @last ||= 0
        super(subclass)
        subclass.uuids = @uuids
        subclass.foreign_ids = @foreign_ids
      end

      # Create a new UUIDObject with unique UUID
      # @return [self]
      def new
        object = super
        @uuids[object.uuid] = object
        @last = object.uuid
        @foreign_ids[object.foreign_id] = object if object.foreign_id
        object
      end

      # Get object by UUID
      # @param uuid
      # @return [self, nil]
      def get(uuid)
        @uuids[uuid] if @uuids[uuid].is_a? self
      end

      # Get object by foreign id
      # @param foreign_id
      # @return [self, nil]
      def get_foreign(foreign_id)
        @foreign_ids[foreign_id] if @foreign_ids[foreign_id].is_a? self
      end

      attr_accessor :uuids, :foreign_ids, :last
    end

    def initialize
      @uuid = __gen_uuid
      @foreign_ids = []
    end

    attr_reader :uuid, :foreign_id

    private

    def __gen_uuid
      newuuid = (Time.now.to_f * 1000).to_i * 10000
      if (self.class.last / 1000) == (newuuid / 1000)
        newuuid += (self.class.last % 1000) + 1
      end
      newuuid
    end
  end

  # PubSub abstract container
  # @abstract
  class PubSub
    def initialize(buffer_size: 1024)
      @buffer_size = buffer_size
      @listeners = []
      @messages = []
    end

    # Push message to all listeners
    # @param msg [UUIDObject]
    # @return [void]
    def push(msg)
      @listeners.each { |x| x.call(msg) }
      @messages.append(msg)
      @messages.shift if @messages.length > @buffer_size
    end

    # Add a listener to the PubSub
    # @param listener [#call]
    # @return [void]
    def listen(listener)
      @listeners.append(listener)
    end

    # Pull all messages since UUID
    # @param uuid [String]
    # @return [Array<UUIDObject>]
    def pull(uuid)
      output = []
      @messages.reverse_each do |x|
        break if x.uuid < uuid

        output.unshift(x)
      end
      output
    end
  end

  # Message struct
  class Message < UUIDObject
    def initialize(datahash, **params)
      super(**params)

      @from = UUIDObject.get(datahash["from"])
      @to = UUIDObject.get(datahash["to"])
      @content = datahash["content"]
      # @reply_to = datahash["reply_to"] # TODO: make this make sense
      @attachments = datahash["attachments"]
    end

    attr_reader :from, :to, :content, :reply_to, :attachments
  end

  # User struct
  class User < UUIDObject
    def initialize(datahash, **params)
      super(**params)

      @id = datahash["id"]
end
