# frozen_string_literal: true

require_relative '.env'
require_relative 'proto'
require 'landline'
require 'json'

# Primary server class
class HeimdallServer < Landline::App
  before do
    # Match data type against a list of datatypes
    # @param obj [Object]
    # @param type [Array, Class]
    def match_type(obj, type)
      if type.is_a? Array
        type.any? { |t| obj.is_a? t }
      else
        obj.is_a? type
      end
    end

    # Validate json body for a post request
    # @param args [Hash] hash of key - type pairs to check JSON data against
    def validate_json(**args)
      die(400, backtrace: ['JSON body expected']) unless json?
      data = begin
        JSON.parse(request.body)
      rescue StandardError
        die(400, backtrace: ['JSON body is invalid'])
      end
      args.each do |k, v|
        unless data.include?(k) and match_type(data[k], v)
          die(400, backtrace: ["Key #{k} is missing"])
        end
      end
    end
  end

  filter do
    request.cookies["token"] == BOT_TOKEN
  end

  path "/user" do
    post "/register" do
      validate_json("id" => Integer,
                    "username" => String,
                    "nickname" => [String, NilClass])
      
    end
  end

  handle do |status, backtrace: nil|
    page = JSON.dump({
                       "error" => backtrace.join("\n"),
                       "code" => status
                     })
    [{
      "content-length": page.bytesize,
      "content-type": "application/json"
    }, page]
  end
end

run HeimdallServer.new
